#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
File Name: sequence2structure.py
Author: Dingfeng Wu
Creator: Dingfeng Wu
Date Created: 2024-01-03
Last Modified: 2023-01-05
Version: 1.0.0
License: MIT License
Description: Sequence-to-Structure (Seq2struc) is a computing process based on RDkit and the characteristics of cyclic peptide sequences, which can create cyclic peptide sequecne and convert sequence to cyclic peptide SMILES.

Copyright Information: Copyright (c) 2023 dfwlab (https://dfwlab.github.io/)

The code in this script can be used under the MIT License.
"""

import pandas as pd
from IPython.display import SVG, display
from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
import re
from .setting import *

def seq2stru_essentialAA(sequence, cyclic=True):

    """
    ``"cyclic"`` controls whether it is looped, and when ``"cyclic = True"``, it generates a looped structure, otherwise a chain.
    
    Exmaple1::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='APG', cyclic=False)
        print('SMILES:', smiles)
        basic.plot_molecule(peptide, w=300, h=300, isdisplay=True)

    SMILES: C[C@H](N)C(=O)N1CCC[C@H]1C(=O)NCC(=O)O
    
    .. image:: s2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%


    :py:func:`basic.plot_molecule` function is used to output the resulting structure, or you can use :py:func:`basic.plot_smiles` .

    Example2::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='APG', cyclic=True)
        print('SMILES:', smiles)
        basic.plot_smiles(smiles, w=300, h=300, isdisplay=True)
        
    SMILES: C[C@H](N)C(=O)N1CCC[C@H]1C(=O)NCC(=O)O

    .. image:: s2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%

    Enter a sequence in 'Ala-Ala-Cys-Asp' format that will be converted to a single-letter structure by the :py:func:`code2symbol` function.

    Example3::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='Ala-Ala-Cys-Asp', cyclic=True)
        print('SMILES:', smiles)
        
    SMILES: C[C@@H]1NC(=O)[C@H](C)NC(=O)[C@H](CC(=O)O)NC(=O)[C@H](CS)NC1=O

    """
    try:
        if '-' in sequence:
            sequence = [code2symbol(code)[0] for code in sequence.split('-')]
            sequence = ''.join(sequence)
        peptide = create_peptide_of_essentialAA(sequence, cyclic=cyclic)
        peptide = Chem.RemoveHs(peptide)  # 移除所有隐式氢原子
        Chem.AssignAtomChiralTagsFromStructure(peptide)  # 重新计算所有隐式氢原子
        smiles = Chem.MolToSmiles(peptide, canonical=True)
        return smiles, peptide
    except:
        return None, None


def seq2stru_no_essentialAA(sequence, references, cyclic=True):
    """
    Example::

        sequence = 'Aad--NMe-Ala--4OH-Thr--3Me-Pro'
        smiles, peptide = sequence2structure.seq2stru_no_essentialAA(sequence, references, cyclic=True)
        print(smiles)
        basic.plot_molecule(peptide, w=300, h=300, isdisplay=True)

    SMILES : CC1CCN2C(=O)C(C(O)CO)NC(=O)C(C)N(C)C(=O)C(CCCC(=O)O)NC(=O)C12

    .. image:: ns2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%

    """
    try:
        sequence = [i.strip() for i in re.sub(r'\(\d+\)', '', sequence).split('--')]
        peptide = create_peptide_of_no_essentialAA(sequence, references, cyclic=cyclic)
        smiles = Chem.MolToSmiles(peptide, canonical=True)
        try:
            peptide = Chem.RemoveHs(peptide)  # 移除所有隐式氢原子
            Chem.AssignAtomChiralTagsFromStructure(peptide)  # 重新计算所有隐式氢原子
            return smiles, peptide
        except:
            return smiles, peptide
    except:
        return None, None


def code2symbol(code):
    """

    :param code: sequence.
    :type code: amino acid chain format
    :return: One letter code

    """

    # 打开文本文件并读取内容
    file_path = AminoAcids_path  # 替换为你的文件路径
    AminoAcids = []
    with open(file_path, 'r', encoding='utf-8') as file:
        """Reads the contents of the file line by line and adds it to the list"""
        for line in file:
            """# 去除行尾的换行符，并将参数添加到列表中"""
            AminoAcids.append(eval(line.strip()))
    c2s = {i[1]: [i[0], i[2]] for i in AminoAcids}
    if code.upper() in c2s.keys():
        return c2s[code.upper()]
    else:
        return None


def link_aa_by_peptide_bond(mol, c_index, n_index):
    """
    Link two amino acids by a peptide bond in a given molecule.

    :param mol: The molecule object representing the amino acids to be linked.
    :type mol: Chem.Mol
    :param c_index: The index of the carbon atom in the molecule where the peptide bond formation will start (usually the carbonyl carbon of one amino acid).
    :type c_index: int
    :param n_index: The index of the nitrogen atom in the molecule where the peptide bond formation will end (usually the amino nitrogen of another amino acid).
    :type n_index: int
    :return: The modified molecule object with the two amino acids linked by a peptide bond.
    :rtype: Chem.Mol

    This function first tries to identify the oxygen and hydrogen atoms of the hydroxyl group attached to the specified carbon atom (`c_index`). It does this by iterating through the neighbors of the carbon atom and looking for an oxygen atom with a single bond to the carbon. Then, it finds the hydrogen atom attached to that oxygen.

    Next, it creates an editable version of the molecule (`emol`) using `Chem.EditableMol`. It then removes the identified hydrogen and oxygen atoms (if they were found) from the editable molecule.

    Finally, it adds a single bond between the specified carbon (`c_index`) and nitrogen (`n_index`) atoms to form the peptide bond. The modified molecule is then retrieved from the editable molecule and returned.
    """

    o_index = None
    h_index = None
    # 查找与该碳原子相连的羟基的氧和氢原子
    for atom in mol.GetAtomWithIdx(c_index).GetNeighbors():
        if atom.GetSymbol() == 'O' and mol.GetBondBetweenAtoms(atom.GetIdx(), c_index).GetBondType() == Chem.BondType.SINGLE:
            o_index = atom.GetIdx()
            for h_atom in atom.GetNeighbors():
                if h_atom.GetAtomicNum() == 1:  # 氢原子
                    h_index = h_atom.GetIdx()
    # print(o_index, h_index)
    # 创建一个可编辑的分子
    emol = Chem.EditableMol(mol)
    # 然后再删除羟基氧原子和氢原子
    emol.RemoveAtom(h_index) if h_index else 1
    emol.RemoveAtom(o_index) if o_index else 1
    emol.AddBond(c_index, n_index, order=Chem.rdchem.BondType.SINGLE)
    # 获取修改后的分子
    mol = emol.GetMol()
    return mol

def detect_backbone(mol):
    """
    Detect the backbone of a given molecule.

    :param mol: The molecule object for which the backbone is to be detected.
    :type mol: Chem.Mol
    :return: A tuple containing two elements:

    The first element is the backbone of the molecule as a tuple of atom indices if matches are found; otherwise, an empty string.
    The second element is a list of the backbone atom indices in reverse order if matches are found; otherwise, None.
    
    :rtype: tuple[str, list[int] or None]

    This function aims to identify the backbone of a given molecule. 

    """

    # 识别骨架并将骨架分子编号为0至N，其他侧链分子编号大于N
    # renumber the backbone atoms so the sequence order is correct:
    # N端乙酰化和C端酰胺化: https://zhuanlan.zhihu.com/p/540769025
    # 肽链一般从N端读到C端,C端一般保留一个COOH,因此C端为HO-C(=O),中间一个C,然后加一个肽键NC(=O),后续就都是C和肽键，一直到N端
    # 正反识别都可以，反向就是OC(=O)CNC(=O)CNC(=O)CN...,如果C端酰胺化,则为NC(=O)CNC(=O)CNC(=O)CN...
    # N端不管有没有乙酰化,都会保留N,无乙酰化为-NH2,有乙酰化为-NC(=O)CH3;
    bbsmiles = "C(=O)CN" * len(
        mol.GetSubstructMatches(Chem.MolFromSmiles('NCC=O')))  # 使用GLY最小氨基酸单元识别氨基酸数量，生成骨架generate backbone SMILES
    # backbone = mol.GetSubstructMatches(Chem.MolFromSmiles(bbsmiles))[0] 报错tuple index out of range
    matches = mol.GetSubstructMatches(Chem.MolFromSmiles(bbsmiles))
    backbone = ''
    if matches:  # 检查是否有匹配结果
        backbone = matches[0]
        # 其他操作
    else:
        print("No matches found.")
        return backbone, None # 如果找不到肽键，或者数量不同则退出，比如存在两个线性肽链通过两个链接形成环的情况，需要逐一搜索；
    backbone_idx = list(backbone)
    backbone_idx.reverse()
    return backbone, backbone_idx
 

def create_peptide_of_essentialAA(sequence, cyclic=True):
    """

    :param sequence: sequence of essential AA.
    :type sequence: One letter code
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """

    try:
        mol = Chem.MolFromSequence(sequence)
    except:
        return None
    if cyclic:
        backbone, backbone_idx = detect_backbone(mol)
        c_index = backbone[0]
        n_index = backbone[-1]
        mol = link_aa_by_peptide_bond(mol, c_index, n_index)
    return mol


def reference_aa_monomer(monomers_path): 
    """

    :param monomers_path: sequence of essential AA.
    :type sequence: One letter code
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """
    if monomers_path:
        monomers = pd.read_csv(monomers_path, sep='\t', index_col=0)
    else:
        monomers = pd.read_csv(monomer_path, sep='\t', index_col=0)
    references = {}
    for i in monomers.index:
        smile = monomers.loc[i, 'Smiles']
        code = monomers.loc[i, 'Code']
        weight = float(monomers.loc[i, 'Weight'])
        symbol = monomers.loc[i, 'Symbol']
        symbol = code if str(symbol).strip() == '' or str(symbol).upper().strip() == 'NAN' else symbol
        try:
            aa = Chem.MolFromSmiles(smile)
            match = aa.GetSubstructMatches(Chem.MolFromSmiles("NCC=O"))
            if match and match[0]:
                num_atoms = aa.GetNumAtoms()
                references[code] = [aa, symbol, num_atoms, match[0]]
        except:
            pass
    return references

def create_peptide_of_no_essentialAA(sequence, references, cyclic=True):
    """

    :param sequence: sequence of noessential AA.
    :type sequence: One letter code
    :param references: monomers.
    :type references: from :py:func:`reference_aa_monomer`
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """
    peptide, pep_idx = references[sequence[0]][0], references[sequence[0]][-1]
    pep_c_index, first_n_index = pep_idx[2], pep_idx[0]
    for aa_code in sequence[1:]:
        aa, idx = references[aa_code][0], references[aa_code][-1]
        c_index, n_index = idx[2], idx[0]
        peptide, pep_c_index = connect_two_aa_with_peptide_bond(peptide, pep_c_index, aa, c_index, n_index)
    if cyclic:
        peptide = link_aa_by_peptide_bond(peptide, pep_c_index, first_n_index)
    return peptide

def connect_two_aa_with_peptide_bond(aa1, c_index1, aa2, c_index2, n_index2):
    """
    Connect two amino acids with a peptide bond.

    :param aa1: The first amino acid molecule.
    :type aa1: Chem.Mol
    :param c_index1: The index of the carbon atom in the first amino acid where the peptide bond will be formed.
    :type c_index1: int
    :param aa2: The second amino acid molecule.
    :type aa2: Chem.Mol
    :param c_index2: The index of the carbon atom in the second amino acid which will be involved in the peptide bond formation.
    :type c_index2: int
    :param n_index2: The index of the nitrogen atom in the second amino acid where the peptide bond will be formed.
    :type n_index2: int
    :return: A tuple containing two elements:
        - The connected peptide molecule formed by joining the two amino acids with a peptide bond.
        - The updated index of the carbon atom in the second amino acid (after accounting for the combination of the two amino acids).
    :rtype: tuple[Chem.Mol, int]

    """
    c_index2, n_index2 = c_index2 + aa1.GetNumAtoms() - 1, n_index2 + aa1.GetNumAtoms() - 1
    peptide = Chem.CombineMols(aa1, aa2)
    peptide = link_aa_by_peptide_bond(peptide, c_index1, n_index2)
    return peptide, c_index2


def get_molblock(mol):
    """
    Retrieve the MolBlock representation of a given molecule.

    :param mol: The molecule object for which the MolBlock is to be retrieved.
    :type mol: Chem.Mol
    :return: The MolBlock string representation of the molecule. The `forceV3000=True` parameter indicates that the MolBlock should be in the V3000 format (if available and applicable).
    :rtype: str

    """

    return Chem.MolToMolBlock(mol, forceV3000=True)


def get_molblock_from_smiles(smiles):
    """
    Generate the MolBlock representation of a molecule from its SMILES string.

    :param smiles: The SMILES string representing the molecule.
    :type smiles: str
    :return: The MolBlock string representation of the molecule. The `forceV3000=True` indicates that the MolBlock will be in the V3000 format (if available and applicable).
    :rtype: str

    This function first converts the given SMILES string into a molecule object using `Chem.MolFromSmiles`. If the conversion is successful, it then retrieves the MolBlock representation of the molecule using `Chem.MolToMolBlock` with the `forceV3000=True` option to potentially get the MolBlock in the V3000 format. If the conversion from SMILES to molecule fails, it will likely return `None` as `Chem.MolToMolBlock` will be called on a `None` object (since `mol` would be `None` if the `Chem.MolFromSmiles` conversion fails).
    """
    mol = Chem.MolFromSmiles(smiles)
    return Chem.MolToMolBlock(mol, forceV3000=True)



from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import AllChem, Descriptors, rdMolDescriptors, Lipinski, rdFingerprintGenerator, \
    MACCSkeys, RDKFingerprint
from rdkit.Chem.rdMolDescriptors import CalcTPSA

logger = RDLogger.logger()
logger.setLevel(RDLogger.CRITICAL)  # 只显示 CRITICAL 级别的日志

################## 化学性质 ##################
def chemial_physical_properties_from_smiles(smiles):
    """
    Calculate the cyclopeptide chemical property from SMILES.

    :param smiles: Cyclic peptides SMILES information.
    :return: chemial and physical properties
    :rtype: dict{property name: value}

    you can also use :py:func:`cal_chemial_physical_properties` to calculate the properties from mol

    List of the chemical properties:

    .. csv-table:: 
        :file: ../states/Chemical_P.csv
        :header: "Property", "Description"

    """

    try:
        mol = Chem.MolFromSmiles(smiles)
    except:
        return {}
    properties = cal_chemial_physical_properties(mol)
    properties = cal_rules(properties)
    properties['RDKit_Fingerprint'] = cal_RDKit_fingerprint(mol)
    properties['Daylight_like_Fingerprint'] = cal_daylight_like_fingerprint(mol)
    properties['Morgan_Fingerprint'] = cal_morgan_fingerprint(mol)
    properties['MACCS_Keys'] = cal_MACCS_keys(mol)
    return properties



def cal_chemial_physical_properties(mol):
    """
    Calculate chemical and physical properties of a molecule.

    :param mol: An RDKit molecule object for which properties are to be calculated.
    :type mol: rdkit.Chem.rdchem.Mol
    :return: A dictionary containing the calculated properties. Keys are property names and values are the corresponding calculated values.
    :rtype: dict

    This function computes various chemical and physical properties of a given molecule
    using RDKit's descriptor calculators.
    """

    properties = {
        'Exact_Mass': Descriptors.ExactMolWt(mol),
        'Topological_Polar_Surface_Area': CalcTPSA(mol),
        'Complexity': Descriptors.FpDensityMorgan1(mol),
        'Crippen_LogP': Descriptors.MolLogP(mol),  # 计算LogP Crippen
        'Heavy_Atom_Count': Lipinski.HeavyAtomCount(mol),
        'Hydrogen_Bond_Donor_Count': Lipinski.NumHDonors(mol),
        'Hydrogen_Bond_Acceptor_Count': Lipinski.NumHAcceptors(mol),
        'Rotatable_Bond_Count': Lipinski.NumRotatableBonds(mol),
        'Formal_Charge': Chem.GetFormalCharge(mol),
        'Refractivity': Descriptors.MolMR(mol),
        'Number_of_Rings': rdMolDescriptors.CalcNumRings(mol),
        'Number_of_Atoms': mol.GetNumAtoms(),
    }
    return properties



def cal_rules(properties):
    """
    :param properties: A list of calculated properties.
    :type properties: calculated by :py:func:`chemial_physical_properties_from_smiles`
    :return: properties

    .. list-table::
        :header-rows: 1

        * - Property
          - Description
        * - Rule of Five
          - Indicates non-compliance with Lipinski's rule of five, suggesting potential issues with bioavailability as an oral drug.
        * - Veber's Rule
          - Shows non-adherence to Veber's rules, potentially impacting oral bioavailability and permeability.
        * - Ghose Filter
          - A molecular property filter used to assess the drug-likeness of a compound based on its physicochemical properties.

     """
    
    # Lipinski 规则五
    lipinski_rule_of_five = properties['Hydrogen_Bond_Donor_Count'] <= 5 and \
                            properties['Hydrogen_Bond_Acceptor_Count'] <= 10 and \
                            properties['Exact_Mass'] <= 500 and \
                            properties['Crippen_LogP'] <= 5

    # Veber 规则
    vebers_rule = properties['Topological_Polar_Surface_Area'] <= 140 and \
                  properties['Rotatable_Bond_Count'] <= 10

    # 检查 Ghose Filter 条件
    ghose_filter = 160 <= properties['Exact_Mass'] <= 480 and \
                   0.4 <= properties['Crippen_LogP'] <= 5.6 and \
                   20 <= properties['Number_of_Atoms'] <= 70

    properties['Rule_of_Five'] = lipinski_rule_of_five
    properties["Veber_Rule"] = vebers_rule
    properties['Ghose_Filter'] = ghose_filter
    return properties


################## 分子指纹 ##################

def cal_RDKit_fingerprint(mol):
    """
    Generate an RDKit fingerprint for a given molecule and return it as a bit vector.

    :param mol: The molecule object for which the RDKit fingerprint is to be generated.
    :type mol: Chem.Mol
    :return: The RDKit fingerprint represented as a bit vector (a string of binary digits).
    :rtype: str

    This function takes a molecule object (`mol`) and uses the `RDKFingerprint` function to create an RDKit fingerprint. Then, it converts this fingerprint into a bit string using the `ToBitString` method. The resulting bit vector is then returned, which can be used for various purposes such as similarity calculations or molecule identification in cheminformatics applications.
    """

    # 生成RDKit指纹
    fp = RDKFingerprint(mol)
    bit_vector = fp.ToBitString()
    return bit_vector

def cal_daylight_like_fingerprint(mol):
    """
    Generates a Daylight-type topological fingerprint from a given molecule.
        
    Parameters Mol:
    Mol: RDKit Mol object representing the input molecule.
        
    Return value bit_vector:
    bit_vector: STR type, a Bit string representation of a molecule's topological fingerprint.
    """

    # 生成拓扑指纹
    fp = rdFingerprintGenerator.GetFPs([mol])
    bit_vector = fp[0].ToBitString()
    return bit_vector


def cal_morgan_fingerprint(mol):
    """
    Generate a Morgan fingerprint (radius 2)
    
    This function takes a molecular object as input and generates its Morgan fingerprint.
    The Morgan fingerprint is a type of molecular descriptor used to encode the structural information of a molecule.
    Here, a Morgan fingerprint with a radius of 2 is used, meaning it considers the molecular environment up to 2 bonds away from each atom.
    
    Parameters mol:
    mol - A molecular object, representing the molecule for which to generate the Morgan fingerprint.
    
    Returns bit_vector:
    bit_vector - A string representing the Morgan fingerprint of the molecule.
    """

    # 生成Morgan指纹（半径为2）
    morgan_fp = AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=1024)
    bit_vector = morgan_fp.ToBitString()
    return bit_vector


def cal_MACCS_keys(mol):
    """
    Generate MACCS keys fingerprint for a given molecule.

    This function calculates the MACCS (Molecular ACCess System) keys fingerprint,
    which is a 166-bit structural key descriptor used for molecular similarity searches.

    Parameters:
    mol (rdkit.Chem.rdchem.Mol): An RDKit molecule object for which to generate the MACCS keys fingerprint.

    Returns:
    str: A bit string representation of the MACCS keys fingerprint for the input molecule.
    """
    # 生成 MACCS keys 指纹
    maccs_fp = MACCSkeys.GenMACCSKeys(mol)
    bit_vector = maccs_fp.ToBitString()
    return bit_vector


def calculate_amino_acid_composition(sequence):
    """
    Calculate the composition of amino acids in a given sequence.

    :param sequence: The amino acid sequence for which the composition is to be calculated.
    :type sequence: str
    :return: A list of tuples where each tuple contains an amino acid and its count in the given sequence.
    :rtype: list[tuple[str, int]]

    This function first reads the contents of a text file named 'AminoAcids.txt' (the file path should be adjusted as needed). The file is expected to contain a list of amino acid representations, likely in a specific format. The function reads each line of the file, strips the newline character at the end of each line, and then uses `eval` to convert the string representation back to its original data type (assuming it was originally a tuple or some other data structure). These are then added to the `AminoAcids` list.

    Next, it extracts the first element of each tuple in the `AminoAcids` list to create a list of amino acids (`aa_list`).

    Finally, it iterates through the `aa_list` and for each amino acid, it counts how many times it appears in the given `sequence`. The result is a list of tuples where each tuple contains an amino acid and its count in the sequence. If an amino acid is not present in the sequence, its count will be 0.
    """
    
    # 打开文本文件并读取内容
    file_path = 'AminoAcids.txt'  # 替换为你的文件路径
    AminoAcids = []
    with open(file_path, 'r', encoding='utf-8') as file:
        """Reads the contents of the file line by line and adds it to the list"""
        for line in file:
            """# 去除行尾的换行符，并将参数添加到列表中"""
            AminoAcids.append(eval(line.strip()))

    aa_list = [i[0] for i in AminoAcids]
    counts = [(aa, sequence.count(aa) if aa in sequence else 0) for aa in aa_list]
    return counts





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
File Name: graph_alignment.py
Author: Dingfeng Wu
Creator: Dingfeng Wu
Date Created: 2023-12-27
Last Modified: 2024-01-03
Version: 1.0.0
License: MIT License
Description: The alignment algorithm is developed by CyclicPepedia based on networkx extension, which can convert cyclic peptides into graphical structures and align them accordingly.

Copyright Information: Copyright (c) 2023 dfwlab (https://dfwlab.github.io/)

The code in this script can be used under the MIT License.
"""
import re
from collections import Counter
from itertools import combinations

import networkx as nx
from networkx.algorithms import isomorphism


def sequence_to_node_edge(sequence):
    """
    Example::

        ref_seq = 'Cys(1)(2)--Cys--OH-DL-Val(2)--4OH-Leu--OH-Ile(1)'

        ref_nodes, ref_edges = ga.sequence_to_node_edge(ref_seq)

    :param sequence: amino acid chain format.
    :type sequence: Format conversion by :py:func:`SequenceTransformer.read_sequence`
    :return: Sequence nodes and edges information
    :rtype: Display with :py:func:`create_graph`

    """
    amino_acids = sequence.split('--')
    nodes = []  # 节点信息: [(0, 'Cys'), (1, 'Cys'), ...]
    edges = []  # 边信息: [(0, 1), (1, 2), ...]
    # 用于存储特殊连接信息
    special_connections = {}
    for i, amino_acid in enumerate(amino_acids):
        # 提取氨基酸名称和特殊连接信息
        amino_acid_name = re.sub(r"\(\d+\)", "", amino_acid)
        # print(amino_acid_name)
        if '-' in amino_acid_name:
            items = [k.strip() for k in amino_acid_name.split('-')]
            for j in range(len(items) - 1):
                nodes.append((str(i) + '_' + str(j), items[j].strip().upper()))
                edges.append((str(i) + '_' + str(j), i))
            amino_acid_name = items[-1]
        nodes.append((i, amino_acid_name.strip().capitalize()))
        # 记录普通连接
        if i > 0:
            edges.append((i - 1, i))
        # 处理特殊连接
        special_conn_ids = re.findall(r"\((\d+)\)", amino_acid)
        # print(special_conn_ids)
        for conn_id in special_conn_ids:
            conn_id = int(conn_id) - 1  # 转换为从0开始的索引
            # 存储特殊连接信息
            if conn_id in special_connections:
                edges.append((special_connections[conn_id], i))
            else:
                special_connections[conn_id] = i
    return nodes, edges


def create_graph(nodes, edges):
    """
    Create a graph using the provided nodes and edges.

    :param nodes: A list of tuples, where each tuple contains a node identifier and its corresponding code.
    :type nodes: list[tuple]
    :param edges: A list of tuples representing the edges in the graph. Each tuple contains two node identifiers indicating the connection between nodes.
    :type edges: list[tuple]
    :return: A networkx graph object (`nx.Graph`) constructed with the given nodes and edges.
    :rtype: nx.Graph

    """
    G = nx.Graph()
    for node, code in nodes:
        G.add_node(node, code=code)
    G.add_edges_from(edges)
    return G


def mcs2similarity(ref, mcs):
    """
    Calculate the similarity between a reference graph and the maximum common subgraph (MCS).

    :param ref: The reference graph used for comparison.
    :type ref: nx.Graph
    :param mcs: The maximum common subgraph (MCS) identified between the reference graph and another graph.
    :type mcs: nx.Graph
    :return: The similarity value between the reference graph and the MCS, calculated as the ratio of the total number of nodes and edges in the MCS to the total number of nodes and edges in the reference graph.
    :rtype: float

    This function computes the similarity between a given reference graph (`ref`) and the maximum common subgraph (`mcs`). The similarity is determined by taking the sum of the number of nodes and edges in the MCS and dividing it by the sum of the number of nodes and edges in the reference graph. The resulting value is a float that represents the degree of similarity between the two graphs.
    """

    return float(len(mcs.nodes) + len(mcs.edges)) / (len(ref.nodes) + len(ref.edges))


def mcs_similarity(query, reference):
    """
    Calculate the similarity between a query graph and a reference graph based on the maximum common subgraph (MCS).

    :param query: The query graph for which similarity with the reference graph is to be calculated.
    :type query: nx.Graph
    :param reference: The reference graph to compare against the query graph.
    :type reference: nx.Graph
    :return: A tuple containing two values. The first value is the total number of nodes and edges in the maximum common subgraph (MCS) if it exists; otherwise, it's 0. The second value is the similarity score between the reference graph and the MCS calculated using the `mcs2similarity` function if the MCS exists; otherwise, it's 0.0.
    :rtype: tuple[int, float]

    """

    subgraphs = generate_subgraphs_from_edges(query)
    n, subg, max_subg = find_max_match_subgraph(subgraphs, reference)
    if max_subg:
        return len(max_subg.nodes) + len(max_subg.edges), mcs2similarity(reference, max_subg)
    return 0, 0.0


def graph_similarity(query, reference):
    """

    :param query: amino acid chain format.
    :type query: Format conversion by :py:func:`SequenceTransformer.read_sequence`
    :return: Sequence nodes and edges information
    :rtype: Display with :py:func:`create_graph`

    """
    distance = nx.graph_edit_distance(query, reference, node_match=node_match)
    similarity = 1 - distance / float(len(query.nodes) + len(query.edges) + len(reference.nodes) + len(reference.edges))
    return distance, similarity


def node_match(n1, n2):
    """Node matching function that matches only if the node's code attribute is the same."""
    # print(n1['code'], n2['code'], n1['code'] == n2['code'])
    return n1['code'] == n2['code']


def generate_subgraphs_from_edges(query_G):
    """Generates a subgraph using the given edge set and finds all the connected components."""
    subgraphs = {}
    for n in range(1, len(query_G.edges) + 1):
        for sub_edges in combinations(query_G.edges(), n):
            # Create a subgraph with the selected edges
            subg = query_G.edge_subgraph(sub_edges).copy()
            if nx.is_connected(subg) and subg.edges():
                subgraphs[len(subg.nodes)] = subgraphs.get(len(subg.nodes), [])
                subgraphs[len(subg.nodes)].append(subg)
    return subgraphs


def is_subgraph_of(query, ref):
    """
    Check if a given query graph is a subgraph of a reference graph.

    :param query: The graph that is being checked to see if it is a subgraph of the reference graph.
    :type query: nx.Graph
    :param ref: The reference graph against which the query graph is being compared.
    :type ref: nx.Graph
    :return: True if the query graph is a subgraph of the reference graph (i.e., there is a subgraph in the reference graph that is isomorphic to the query graph), otherwise False.
    :rtype: bool

    """

    GM = isomorphism.GraphMatcher(ref, query, node_match=node_match)
    return GM.subgraph_is_isomorphic()


def find_max_match_subgraph(subgraphs, ref_G):
    """
    Find the maximum match subgraph among a collection of subgraphs with respect to a reference graph.

    :param subgraphs: A dictionary where the keys are some identifiers (presumably related to the subgraphs) and the values are lists of subgraphs. Each subgraph is expected to be of type `nx.Graph`.
    :type subgraphs: dict
    :param ref_G: The reference graph against which the subgraphs are being compared to find the maximum match.
    :type ref_G: nx.Graph
    :return: A tuple containing three elements:
        - The identifier `n` corresponding to the group of subgraphs where the maximum match subgraph was found (if any). If no match is found, this could be an arbitrary value from the `sorted(subgraphs.keys())` iteration.
        - The subgraph `subg` that was being checked when the maximum match subgraph was found (if any). If no match is found, this could be an arbitrary subgraph from the last iteration.
        - The maximum match subgraph `max_subg` if it exists; otherwise, `None`. If a match is found, it will be of type `nx.Graph`.
    :rtype: tuple

    """

    max_subg = None
    for n in sorted(subgraphs.keys()):
        # print(n, len(subgraphs[n]))
        for subg in subgraphs[n]:
            if is_subgraph_of(subg, ref_G):
                max_subg = subg
                break
        else:
            break
    return n, subg, max_subg


def amino_acid_composition_freq(query, ref):
    """
    Calculate the similarity between the amino acid composition frequencies of two sequences (query and reference).

    :param query: The query sequence for which the amino acid composition frequency is to be analyzed.
    :type query: sequence-like (e.g., list, tuple, string)
    :param ref: The reference sequence for comparison.
    :type ref: sequence-like (e.g., list, tuple, string)
    :return: The similarity value between the amino acid composition frequencies of the query and reference sequences. The similarity is calculated as the ratio of the minimum similarity to the maximum similarity. If the maximum similarity is zero, the returned similarity value is zero.
    :rtype: float

    """

    query_freq = Counter(query)
    ref_freq = Counter(ref)
    query_freq_norm = {aa: count / float(len(query_freq)) for aa, count in query_freq.items()}
    ref_freq_norm = {aa: count / float(len(ref_freq)) for aa, count in ref_freq.items()}
    all_amino_acids = set(query_freq_norm.keys()).union(set(ref_freq_norm.keys()))
    # Calculate similarity
    min_similarity = sum(min(query_freq_norm.get(aa, 0), ref_freq_norm.get(aa, 0)) for aa in all_amino_acids)
    max_similarity = sum(max(query_freq_norm.get(aa, 0), ref_freq_norm.get(aa, 0)) for aa in all_amino_acids)
    similarity = min_similarity / max_similarity if max_similarity > 0 else 0
    return similarity



from rdkit import Chem
from rdkit.Chem import AllChem

################## 结构格式互转 ##################
# 自动识别并读人SMILES InChI Molblock SDFblock PDBblock
# https://chemistry.stackexchange.com/questions/34563/pubchem-inchi-smiles-and-uniqueness
# 以InChi=1S/…开头的InChi标识符是标准InChI。在标准InChI中，InChI标识符“对于任何移动氢原子的排列都必须相同”
# 以InChI=1/…开始为非标准InChI，包括一个以/f开头的额外层（fixed-hydrogen layer）。
# 标准InChI生成结构与SMILES不同，而非标准InChI生成结构与SMILES相同，非标准InChI生成生成代码：Chem.MolToInchi(mol1, options='/FixedH')
# 自动识别并读人SMILES InChI Molblock SDFblock PDBblock


# 生成 SMILES InChI InChIKey Molblock SDFblock PDBblock
def output_molecule(mol, pdbblock=None, conformation=None):
    """

    :param mol: mol file.
    :param pdbblock: Optional "conformation".
    :type pdbblock: pdbblock or None
    :param conformation: Optional "conformation".
    :type conformation: conformation or None
    :return: mol

    """

    smiles = Chem.MolToSmiles(mol, canonical=True, isomericSmiles=True)
    inchi = Chem.MolToInchi(mol, options='/FixedH')
    inchikey = Chem.InchiToInchiKey(inchi)
    molblock = Chem.MolToMolBlock(mol, includeStereo=True)
    # PDBblock需要通过空间结构保存手性, 因此其他格式转PDB时需要先创建3D构象;
    if pdbblock is None and conformation is not None:
        pdbblock = Chem.MolToPDBBlock(conformation)
    return {'smiles': smiles, 'inchi': inchi, 'inchikey': inchikey, 'molblock': molblock, 'pdbblock': pdbblock}


def predict_3d_conformation(mol):
    """
    Predict the 3D structure of a molecule.

    :param mol: The input molecule for which the 3D conformation is to be predicted.
    :type mol: Chem.Mol
    :return: The molecule with predicted 3D conformation, including added hydrogens and generated 3D coordinates.
    :rtype: Chem.Mol

    This function takes an input molecule and predicts its 3D structure. First, it creates a copy of the input molecule and adds hydrogens to it using `Chem.AddHs`. Then, it generates 3D coordinates for the molecule with added hydrogens by employing `AllChem.EmbedMolecule` with the `AllChem.ETKDG()` method. Finally, the function returns the molecule with the predicted 3D conformation.

    """
    # molecule = Chem.MolFromSmiles(smiles)
    mol_3d = Chem.AddHs(mol.__copy__())  # 添加氢
    AllChem.EmbedMolecule(mol_3d, AllChem.ETKDG())  # 生成 3D 坐标
    return mol_3d


def mol_optimize(mol):
    """
    Optimizes the 3D structure of a molecule using the Universal Force Field (UFF).

    :param mol: The molecule to be optimized. This molecule should have 3D coordinates.
    :type mol: Chem.Mol
    :return: The optimized molecule with updated 3D coordinates.
    :rtype: Chem.Mol

    This function utilizes the Universal Force Field (UFF) to optimize the 3D structure of a given molecule. The input molecule must possess 3D coordinates for the optimization process to be carried out effectively. Once the optimization is complete, the function returns the molecule with its 3D coordinates updated to reflect the optimized structure.

    """

    # 能量最小化：UFF (Universal Force Field)适用于小分子的力场；多肽建模和优化，专门的生物分子建模工具（如 AMBER、CHARMM 或 GROMACS）会更合适
    AllChem.UFFOptimizeMolecule(mol)  # 使用 UFF 力场进行优化
    return mol



def mol2molblock(mol):
    """
    Convert a molecule object to a MolBlock string.

    :param mol: The molecule object to be converted.
    :type mol: Chem.Mol
    :return: The MolBlock string representation of the input molecule.
    :rtype: str

    """

    return Chem.MolToMolBlock(mol)


def mol2pdbblock(mol):
    """
    Convert a molecule object to a PDBBlock string.

    :param mol: The molecule object to be converted.
    :type mol: Chem.Mol
    :return: The PDBBlock string representation of the input molecule.
    :rtype: str

    """

    return Chem.MolToPDBBlock(mol)



from rdkit import Chem
import os
import pandas as pd
from rdkit.Chem.Draw import rdMolDraw2D
from IPython.display import SVG, display
import networkx as nx
import matplotlib.pyplot as plt
import io

###############结构输出####################


def save_mol(molblock, output_file):
    """
    Save a molecule represented by a MolBlock string to a file.

    :param molblock: The MolBlock string representation of the molecule to be saved.
    :type molblock: str
    :param output_file: The name of the file to which the molecule will be saved.
    :type output_file: str
    :return: molblock file

    This function takes a MolBlock string (molblock) and saves the corresponding molecule to a file with the specified output_file name. 

    """

    mol = Chem.MolFromMolBlock(molblock)
    if mol is not None:
        Chem.MolToMolFile(mol, output_file)


def save_pdb(pdbblock, output_file):
    """
    Save a molecule represented by a PDBBlock string to a file.

    :param pdbblock: The PDBBlock string representation of the molecule to be saved.
    :type pdbblock: str
    :param output_file: The name of the file to which the molecule will be saved.
    :type output_file: str
    :return: pdbblock file

    This function takes a PDBBlock string (`pdbblock`) and saves the corresponding molecule to a file with the specified `output_file` name.
    """

    mol = Chem.MolFromPDBBlock(pdbblock)
    if mol is not None:
        Chem.MolToPDBFile(mol, output_file)

def save_svg(svg, output_file):
    """
    Save an SVG (Scalable Vector Graphics) string to a file.

    :param svg: The SVG string to be saved.
    :type svg: str
    :param output_file: The name of the file to which the SVG will be saved.
    :type output_file: str
    :return: svg file

    This function takes an SVG string and saves it to a file with the specified output_file name.
    """

    with open(output_file, 'w') as f:
        f.write(svg)

###################图像输出#################
def plot_smiles(smiles, output_file='output.svg', w=600, h=600, isdisplay=False):  # 默认尺寸为600x600
    """
    Plot a molecule represented by its SMILES (Simplified Molecular Input Line Entry System) string and optionally save or display the resulting SVG image.

    :param smiles: The SMILES string representing the molecule to be plotted.
    :type smiles: str
    :param output_file: (Optional) The name of the file to which the SVG image of the molecule will be saved. Defaults to 'output.svg'.
    :type output_file: str
    :param w: (Optional) The width of the SVG image in pixels. Defaults to 600.
    :type w: int
    :param h: (Optional) The height of the SVG image in pixels. Defaults to 600.
    :type h: int
    :param isdisplay: (Optional) A boolean flag indicating whether to display the SVG image. If True, the image will be displayed; if False, only the SVG string will be returned. Defaults to False.
    :type isdisplay: bool
    :return: If `isdisplay` is False, the SVG string representing the plotted molecule is returned. If `isdisplay` is True, nothing is returned as the image is directly displayed.
    :rtype: str or None

    This function first attempts to convert the given SMILES string into a molecule object using `Chem.MolFromSmiles`. 

    """
    try:
        m = Chem.MolFromSmiles(smiles)
    except:
        return ''
    # 设置绘图选项，调整分子的大小
    d2d = rdMolDraw2D.MolDraw2DSVG(w, h)
    # 绘制分子
    d2d.DrawMolecule(m)
    d2d.FinishDrawing()
    # 显示图像
    svg = d2d.GetDrawingText()
    if output_file:
        with open(output_file, 'w') as f:
            f.write(svg)
    if isdisplay:
        display(SVG(svg.replace('svg:', '')))
    else:
        return svg


def plot_graph(G, output_file='output.pdf', dpi=60):
    """
    Plot a graph object and optionally save the plot as a PDF file.

    :param G: The graph object to be plotted, typically a networkx graph (e.g., `nx.Graph`).
    :type G: nx.Graph
    :param output_file: (Optional) The name of the file to which the plot will be saved. Defaults to 'output.pdf'.
    :type output_file: str
    :param dpi: (Optional) The dots per inch (resolution) for the saved plot. Defaults to 60.
    :type dpi: int
    :return: pdf file
    
    """

    plt.figure(figsize=(6, 6), dpi=dpi)
    pos = nx.spring_layout(G)  # 生成布局
    nx.draw(G, pos, with_labels=False, node_color='lightblue', node_size=800)
    labels = nx.get_node_attributes(G, 'code')
    # 调整标签位置
    label_pos = {k: [v[0], v[1] + 0] for k, v in pos.items()}
    nx.draw_networkx_labels(G, label_pos, labels=labels)
    if output_file:
        plt.savefig(output_file, format="pdf")
    plt.show()
    

def graph2svg(G, output_file='outpot.svg', dpi=60):
    """
    Convert a graph object to an SVG file and optionally save it.

    :param G: The graph object to be converted, typically a networkx graph (e.g., `nx.Graph`).
    :type G: nx.Graph
    :param output_file: (Optional) The name of the file to which the SVG will be saved. Defaults to 'outpot.svg'.
    :type output_file: str
    :param dpi: (Optional) The dots per inch (resolution) of the graph layout. Defaults to 60.
    :type dpi: int
    :return: svg file

    """

    plt.figure(figsize=(6, 6), dpi=dpi)
    pos = nx.spring_layout(G)  # 生成布局
    svg_buffer = io.StringIO()
    nx.draw(G, pos, with_labels=False, node_color='lightblue', node_size=800)
    labels = nx.get_node_attributes(G, 'code')
    # 调整标签位置
    label_pos = {k: [v[0], v[1] + 0] for k, v in pos.items()}
    nx.draw_networkx_labels(G, label_pos, labels=labels)
    plt.savefig(svg_buffer, format="svg", bbox_inches='tight')
    plt.show()
    if output_file:
        svg = svg_buffer.getvalue()
        with open(output_file, 'w') as f:
            f.write(svg)


import csv
import itertools


def generate_similar_sequences(sequence, replacement_rules_file=None, Replacement_ratio=None):
    """
    Generate similar sequences based on a given sequence and optional replacement rules.

    :param sequence: The input sequence for which similar sequences are to be generated.
    :type sequence: str
    :param replacement_rules_file: (Optional) The file path containing the replacement rules. If not provided, default replacement rules will be used.
    :type replacement_rules_file: str or None
    :param Replacement_ratio: (Optional) The ratio of amino acids to be replaced. If not provided, it will be set to one-third of the total amino acid count.
    :type Replacement_ratio: float or None
    :return: A list of possible similar sequences generated based on the input sequence, replacement rules (either default or from file), and the replacement ratio (either default or specified).
    :rtype: list[str]

    This function generates similar sequences to the given input sequence. 
    """

    # 如果没有指定外部文件，则使用默认规则
    if replacement_rules_file is None:
        replacement_rules = {
            'Glu': ['Asp'],  # 酸性氨基酸
            'Asp': ['Glu'],
            'Lys': ['Arg', 'His'],  # 碱性氨基酸
            'Arg': ['Lys', 'His'],
            'His': ['Lys', 'Arg'],
            'Leu': ['Ile', 'Val', 'Ala'],  # 疏水性氨基酸
            'Ile': ['Leu', 'Val', 'Ala'],
            'Val': ['Leu', 'Ile', 'Ala'],
            'Ala': ['Leu', 'Ile', 'Val'],
            'Phe': ['Tyr', 'Trp'],  # 芳香性氨基酸
            'Tyr': ['Phe', 'Trp'],
            'Trp': ['Phe', 'Tyr'],
            'Ser': ['Thr'],  # 极性氨基酸
            'Thr': ['Ser'],
            'Gly': ['Ala'],  # 小体积氨基酸
            'Cys': ['Ser'],  # 特殊功能基团（极性替换）
            'Met': ['Leu', 'Ile'],  # 疏水/非极性氨基酸
            'Asn': ['Gln'],  # 带酰胺基的极性氨基酸
            'Gln': ['Asn'],
        }
    else:
        replacement_rules = read_replacement_rules_from_file(replacement_rules_file)
    total_aa_count = len(sequence.split('--'))
    # 如果没有指定最大替换数量，则设置为总氨基酸数的三分之一
    if Replacement_ratio is None:
        max_replacements = total_aa_count // 3
    else:
        max_replacements = total_aa_count * Replacement_ratio

    # 使用正则表达式匹配氨基酸和其修饰部分（这里暂不处理修饰部分的特殊逻辑）
    fragments = sequence.split('--')
    all_combinations = []

    # 遍历每一个片段，生成替换选项
    for fragment in fragments:
        aa = fragment
        options = []
        if aa in replacement_rules and replacement_rules[aa]:
            # 随机替换的情况，包括不替换和替换的选项
            options.append(aa)
            for replacement in replacement_rules[aa]:
                options.append(replacement)
        else:
            options.append(aa)

        all_combinations.append(options)

    # 生成所有可能的替换组合
    possible_sequences = []
    for combo in itertools.product(*all_combinations):
        replacements_count = sum(1 for i in range(len(fragments)) if combo[i]!= fragments[i])
        if replacements_count >= 0 and replacements_count <= max_replacements:
            new_sequence = '--'.join(combo)
            possible_sequences.append(new_sequence)

    return possible_sequences


def read_replacement_rules_from_file(file_path):
    """
    Read replacement rules from a specified file.

    :param file_path: The path to the file containing the replacement rules.
    :type file_path: str
    :return: A dictionary where the keys are amino acids and the values are lists of possible replacement amino acids as read from the file.
    :rtype: dict

    This function reads replacement rules from a given file. 
    """

    replacement_rules = {}
    with open(file_path, 'r') as file:
        for line in file.readlines():
            line = line.strip()
            if line:
                key, values_str = line.split(':')
                values = [value.strip() for value in values_str.strip('[]').spanplit(',')]
                replacement_rules[key.strip()] = values
    return replacement_rules


def save_sequences_to_csv(sequences, filename):
    """
    Save a list of sequences to a CSV file.
    
    :param sequences: A list of sequences to be saved to the CSV file.
    :type sequences: list[str]
    :param filename: The name of the CSV file to which the sequences will be saved.
    :type filename: str
    :return: None

    This function takes a list of sequences and saves them to a CSV file with the specified filename. 

    """

    with open(filename, mode='w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(['Sequence'])  # 写入标题
        for sequence in sequences:
            writer.writerow([sequence])